#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>
#include <stdint.h>
#include <stdlib.h>
#include <math.h>

#define ERR_NO_WAVE_MAPPER 100

#define PREFILL_BUFFS 4
#define NUM_BUFFS 6
#define SAMPLE_RATE 44100
#define SAMPLES_PER_EFFECT 44100
#define SAMPLES_PER_BUFFER 22050
#define BYTES_PER_SAMPLE 2

#define BPM_APPROX 460

#define SAMPLES_PER_GATE ( (int32_t)( (SAMPLE_RATE*60)/(BPM_APPROX) ) )

static double noiseTable[SAMPLES_PER_EFFECT] = {0};

// Exponential falloff function, for drum sounds
double C(double sampleIdx, double falloff) {
	if (sampleIdx < 0) {
		return 1;
	}
	sampleIdx /= SAMPLES_PER_EFFECT;
	if (sampleIdx > 1) {
		return 0;
	}
	return pow(1 - sampleIdx, falloff);
}

void noiseTable_init() {
	for(int i = 0 ; i < SAMPLES_PER_EFFECT ; i += 1) {
		noiseTable[i] = (double)rand() / (double)RAND_MAX;
	}
}

double N(int64_t idx) {
	if (idx < 0) {
		return 0;
	}
	return noiseTable[idx % SAMPLES_PER_EFFECT];
}

//
//
//

static int16_t buffs[NUM_BUFFS][SAMPLES_PER_BUFFER] = {0};
int64_t totalGates     = 0;
int64_t nextBuffPlay   = 0;
int64_t nextBuffRender = 0;

void clipSample(double res, int buffIdx, int buffOffset) {
	int64_t currentVal = buffs[buffIdx % NUM_BUFFS][buffOffset];
	currentVal += (int16_t)((res/2.0) * 32768); // Reduce amplitude by 2x to reduce chance of clamping with multiple samples
	
	// Clamping sounds better than overflowing
	if (currentVal > 32767) {
		currentVal = 32767;
	}
	if (currentVal < -32767) {
		currentVal = -32767;
	}
	
	buffs[buffIdx % NUM_BUFFS][buffOffset] = (int16_t)currentVal;
}

void addGateToBuff(int64_t g) {
	#define ROW(gateExpression, sampleExpression) \
	if ( gateExpression ) { \
		int64_t firstGlobalSampleCt = (g * SAMPLES_PER_GATE); \
		for (int32_t s = 0; s < SAMPLES_PER_EFFECT; ++s) { \
			int64_t currentGlobalSampleCt = (firstGlobalSampleCt + s); \
			int64_t buffIdx = currentGlobalSampleCt / SAMPLES_PER_BUFFER; \
			int64_t buffOffset = currentGlobalSampleCt - (buffIdx * SAMPLES_PER_BUFFER); \
			double res = ( sampleExpression ); \
			clipSample( sampleExpression, buffIdx, buffOffset ); \
		} \
	}
	
	#include "song.inc"
}

void wipeBuffer(int buffIdx) {
	for (int i = 0; i < SAMPLES_PER_BUFFER; ++i) {
		buffs[buffIdx][i] = 0;
	}
}

void renderNextBuffer() {
	int currentBuffRender = nextBuffRender;
	int buffIdx = ++nextBuffRender % NUM_BUFFS;
	
	// Wipe out the /following/ buff before we start
	int followingBuff = (buffIdx + 1) % NUM_BUFFS;
	wipeBuffer(followingBuff);
	
	for(;;) {
		
		int nextGate = totalGates;
		int buffForNextGate = (nextGate * SAMPLES_PER_GATE) / SAMPLES_PER_BUFFER;
		
		if (buffForNextGate > currentBuffRender) {
			break;
		}
		
		addGateToBuff(totalGates);
		totalGates++;
	}
}

//
//
//

static WAVEHDR buffh[NUM_BUFFS] = {0};

void playNextBuffer(HWAVEOUT hWo) {
	int idx = ++nextBuffPlay % NUM_BUFFS;
	
	waveOutWrite(hWo, &buffh[idx], sizeof(WAVEHDR)); // returns immediately
}

void CALLBACK waveout_onEvent(HWAVEOUT hWo, UINT uMsg, DWORD_PTR dwInstance, DWORD_PTR dwParam1, DWORD_PTR dwParam2) {
	renderNextBuffer();
	playNextBuffer(hWo);
}

#ifdef SMALL_VER
void _pei386_runtime_relocator() {}
#endif

#ifdef SMALL_VER
void __main() {
#else
int main(int argc, char** argv) {
#endif
	HWAVEOUT hWo;
	WAVEFORMATEX wfx = {
		.nSamplesPerSec  = SAMPLE_RATE,
		.wBitsPerSample  = BYTES_PER_SAMPLE * 8,
		.nChannels       = 1,
		.cbSize          = 0,
		.wFormatTag      = WAVE_FORMAT_PCM,
		.nBlockAlign     = BYTES_PER_SAMPLE,
		.nAvgBytesPerSec = SAMPLE_RATE * BYTES_PER_SAMPLE
	};
	
	if(waveOutOpen(&hWo, WAVE_MAPPER, &wfx, (DWORD_PTR)waveout_onEvent, 0, CALLBACK_FUNCTION) != MMSYSERR_NOERROR) {
		ExitProcess(ERR_NO_WAVE_MAPPER);
	}
	
	noiseTable_init();
	
	// Set up buffers
	for (int i = 0 ; i < NUM_BUFFS; i += 1) {
		buffh[i].dwBufferLength = SAMPLES_PER_BUFFER*BYTES_PER_SAMPLE;
		buffh[i].lpData = (char*) buffs[i];
		waveOutPrepareHeader(hWo, &buffh[i], sizeof(WAVEHDR));
	}
	
	for (int i = 0; i < PREFILL_BUFFS; i += 1) {
		renderNextBuffer();
	}
	for (int i = 0; i < PREFILL_BUFFS; i += 1) {
		playNextBuffer(hWo);
	}
	
	Sleep(INFINITE);
	
	#ifndef SMALL_VER
	return 0; // UNREACHED
	#endif
}
